from copy import deepcopy

import pandas as pd
from sklearn.utils.validation import check_is_fitted


class GAR:
    """Implementation of the Generalized Auto Regression model.

    Parameters
    ----------
    base_model: object, required
        The model used to make the predictions step by step. This object must have a
        ``fit``and ``predict`` method.

    feed_forward: bool, optional, default: ``False``
        If true, feed-forward the predictions of the time_series_models at training and
        prediction time.

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from giottotime.models import GAR
    >>> from sklearn.ensemble import RandomForestRegressor
    >>> time_index = pd.date_range("2020-01-01", "2020-01-30")
    >>> X = pd.DataFrame(np.random.random((30, 5)), index=time_index)
    >>> y_columns = ["y_1", "y_2", "y_3"]
    >>> y = pd.DataFrame(np.random.random((30, 3)), index=time_index, columns=y_columns)
    >>> X_train, y_train = X[:20], y[:20]
    >>> X_test, y_test = X[20:], y[20:]
    >>> random_forest = RandomForestRegressor()
    >>> gar = GAR(base_model=random_forest)
    >>> gar.fit(X_train, y_train)
    >>> predictions = gar.predict(X_test)
    >>> predictions.shape
    (10, 3)

    """

    def __init__(self, base_model: object, feed_forward: bool = False):
        if not hasattr(base_model, "fit") or not hasattr(base_model, "predict"):
            raise TypeError(
                f"{base_model} must implement both 'fit' " f"and 'predict' methods"
            )

        self.base_model = base_model
        self.feed_forward = feed_forward

    def fit(self, X: pd.DataFrame, y: pd.DataFrame, **kwargs: object) -> "GAR":
        """Fit the GAR model according to the training data.

        Parameters
        ----------
        X: pd.DataFrame, shape (n_samples, n_features), required
            Features used to fit the model.

        y: pd.DataFrame, shape (n_samples, horizon), required
            Target values to fit on.

        kwargs: dict, optional
            Optional parameters to be passed to the base model during the fit procedure.

        Returns
        -------
        self: GAR
            The fitted GAR object.

        """
        features = X.copy()
        models_per_predstep = [deepcopy(self.base_model) for _ in range(y.shape[1])]

        for pred_step, model_for_pred_step in enumerate(models_per_predstep, 1):
            target_y = y[f"y_{pred_step}"]
            model_for_pred_step.fit(features, target_y, **kwargs)

            if self.feed_forward:
                predictions = model_for_pred_step.predict(features)
                features[f"preds_{pred_step}"] = predictions

        self.models_per_predstep_ = models_per_predstep
        self.train_features_ = X

        return self

    def predict(self, X: pd.DataFrame) -> pd.DataFrame:
        """Make predictions for each sample and for each prediction step.

        Parameters
        ----------
        X: pd.DataFrame, shape (n_samples, n_features), required
            Features used to predict.

        Returns
        -------
        predictions: pd.DataFrame, shape (n_samples, 1)
            The predictions of the model.

        Raises
        ------
        NotFittedError
            Thrown if the model has not been previously fitted.

        """
        check_is_fitted(self)

        test_features = X.copy()

        predictions = pd.DataFrame(index=test_features.index)

        for pred_step, model_for_pred_step in enumerate(self.models_per_predstep_, 1):
            model_predictions = model_for_pred_step.predict(test_features)
            predictions[f"y_{pred_step}"] = model_predictions

            if self.feed_forward:
                test_features[f"preds_{pred_step}"] = model_predictions

        return predictions



from itertools import product

import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression
from sklearn.utils.validation import check_is_fitted

from giottotime.causality_tests.base import CausalityTest


class ShiftedLinearCoefficient(CausalityTest):
    """Test the shifted linear fit coefficients between two or more time series.

    Parameters
    ----------
    max_shift : int, optional, default: ``10``
        The maximum number of shifts to check for.

    target_col : str, optional, default: ``'y'``
        The column to use as the a reference (i.e., the column which is not
        shifted).

    dropna : bool, optional, default: ``False``
        Determines if the Nan values created by shifting are retained or dropped.

    Examples
    --------

    >>> from giottotime.causality_tests.shifted_linear_coefficient import ShiftedLinearCoefficient
    >>> import pandas.util.testing as testing
    >>> data = testing.makeTimeDataFrame(freq="s")
    >>> slc = ShiftedLinearCoefficient(target_col="A")
    >>> slc.fit(data)
    >>> slc.best_shifts_
    y  A  B  C  D
    x
    A  3  6  8  5
    B  9  9  4  1
    C  8  2  4  9
    D  3  9  4  3
    >>> slc.max_corrs_
    y         A         B         C         D
    x
    A  0.460236  0.420005  0.339370  0.267143
    B  0.177856  0.300350  0.367150  0.550490
    C  0.484860  0.263036  0.456046  0.251342
    D  0.580068  0.344688  0.253626  0.256220
    """

    def __init__(
        self,
        max_shift: int = 10,
        target_col: str = "y",
        dropna: bool = False,
        bootstrap_iterations=1000,
        bootstrap_samples=100,
    ):
        super().__init__(bootstrap_iterations, bootstrap_samples)
        self.max_shift = max_shift
        self.target_col = target_col
        self.dropna = dropna

    def fit(self, data: pd.DataFrame) -> "ShiftedLinearCoefficient":
        """Create the DataFrame of shifts of each time series which maximize the shifted
         linear fit coefficients.

        Parameters
        ----------
        data : pd.DataFrame, shape (n_samples, n_time_series), required
            The DataFrame containing the time-series on which to compute the shifted
            linear fit coefficients.

        Returns
        -------
        self : ``ShiftedLinearCoefficient``

        """
        best_shifts = pd.DataFrame(columns=["x", "y", "shift", "max_corr", "p_values"])
        best_shifts = best_shifts.astype(
            {
                "x": np.float64,
                "y": np.float64,
                "shift": np.int64,
                "max_corr": np.int64,
                "p_values": np.float64,
            }
        )

        for x, y in product(data.columns, repeat=2):
            res = self._get_max_coeff_shift(data, self.max_shift, x=x, y=y)
            best_shift = res[1]
            max_corr = res[0]
            p_value = self._compute_is_test_significant(data, x, y, best_shift)
            # N = data.shape[0] - max_shift
            best_shifts = best_shifts.append(
                {
                    "x": x,
                    "y": y,
                    "shift": best_shift,
                    "max_corr": max_corr,
                    "p_values": p_value,
                },
                ignore_index=True,
            )

        pivot_best_shifts = pd.pivot_table(
            best_shifts, index=["x"], columns=["y"], values="shift"
        )
        max_corrs = pd.pivot_table(
            best_shifts, index=["x"], columns=["y"], values="max_corr"
        )
        p_values = pd.pivot_table(
            best_shifts, index=["x"], columns=["y"], values="p_values"
        )

        self.best_shifts_ = pivot_best_shifts
        self.max_corrs_ = max_corrs
        self.p_values_ = p_values

        return self

    def transform(self, data: pd.DataFrame) -> pd.DataFrame:
        """Shifts each input time series by the amount which maximizes shifted linear
        fit coefficients with the selected 'y' column.

        Parameters
        ----------
        data : pd.DataFrame, shape (n_samples, n_time_series), required
            The DataFrame containing the time series on which to perform the
            transformation.

        Returns
        -------
        data_t : pd.DataFrame, shape (n_samples, n_time_series)
            The DataFrame (Pivot table) of the shifts which maximize the shifted linear
            fit coefficients between each time series. The shift is indicated in rows.

        """
        check_is_fitted(self)
        data_t = data.copy()

        for col in data_t:
            if col != self.target_col:
                data_t[col] = data_t[col].shift(
                    -self.best_shifts_[col][self.target_col]
                )

        if self.dropna:
            data_t = data_t.dropna()

        return data_t

    def _get_max_coeff_shift(
        self, data: pd.DataFrame, max_shift: int, x: str = "x", y: str = "y"
    ) -> (float, int):
        shifts = pd.DataFrame()

        shifts[x] = data[x]
        shifts[y] = data[y]

        for shift in range(1, max_shift):
            shifts[shift] = data[x].shift(shift)

        shifts = shifts.dropna()
        lf = LinearRegression().fit(
            shifts[range(1, max_shift)].values, shifts[y].values
        )

        q = lf.coef_.max(), np.argmax(lf.coef_) + 1
        return q



from typing import Union, List

import numpy as np
import pandas as pd


def _check_input(y_true: np.ndarray, y_pred: np.ndarray) -> None:
    if len(y_pred) != len(y_true):
        raise ValueError(
            f"The arrays must have the same length, but they "
            f"have length {len(y_pred)} and {len(y_true)}."
        )
    if np.isnan(y_true).any() or np.isnan(y_pred).any():
        raise ValueError(
            "The two arrays should not contain Nan values, but they are "
            f"{y_true}, {y_pred}."
        )

    if np.isinf(y_true).any() or np.isinf(y_pred).any():
        raise ValueError(
            "The two arrays should not contain Inf values, but they are "
            f"{y_true}, {y_pred}."
        )


def _convert_to_ndarray(
    y_true: Union[pd.DataFrame, List, np.ndarray],
    y_pred: Union[pd.DataFrame, List, np.ndarray],
):
    if isinstance(y_true, pd.DataFrame):
        y_true = y_true.values
    elif isinstance(y_true, List):
        y_true = np.array(y_true)

    if isinstance(y_pred, pd.DataFrame):
        y_pred = y_pred.values
    elif isinstance(y_pred, List):
        y_pred = np.array(y_pred)

    return y_true, y_pred


def smape(
    y_true: Union[pd.DataFrame, List, np.ndarray],
    y_pred: Union[pd.DataFrame, List, np.ndarray],
) -> float:
    """Compute the 'Symmetric mean absolute percentage error' (SMAPE) between two
    vectors.

    Parameters
    ----------
    y_true : Union[pd.DataFrame, List, np.ndarray], shape (length, 1), required
        The first vector.

    y_pred : Union[pd.DataFrame, List, np.ndarray], shape (length, 1), required
        The second vector.

    Returns
    -------
    smape : float
        The smape between the two input vectors.

    Examples
    --------
    >>> from giottotime.loss_functions import smape
    >>> y_true = [0, 1, 2, 3, 4, 5]
    >>> y_pred = [1.1, 2.3, 0.4, 3.9, 3.1, 4.6]
    >>> smape(y_true, y_pred)
    0.7864893577539014

    """
    y_true, y_pred = _convert_to_ndarray(y_true, y_pred)
    _check_input(y_true, y_pred)

    non_normalized_smape = sum(
        np.abs(y_pred - y_true) / (np.abs(y_pred) + np.abs(y_true))
    )
    non_normalized_smape_filled = np.nan_to_num(non_normalized_smape, nan=0)
    smape = (2 / len(y_pred)) * non_normalized_smape_filled
    return smape


def max_error(
    y_true: Union[pd.DataFrame, List, np.ndarray],
    y_pred: Union[pd.DataFrame, List, np.ndarray],
) -> float:
    """Compute the maximum error between two vectors.

    Parameters
    ----------
    y_true : Union[pd.DataFrame, List, np.ndarray], shape (length, 1), required
        The first vector.

    y_pred : Union[pd.DataFrame, List, np.ndarray], shape (length, 1), required
        The second vector.

    Returns
    -------
    error : float
        The maximum error between the two vectors.

    Examples
    --------
    >>> from giottotime.loss_functions import max_error
    >>> y_true = [0, 1, 2, 3, 4, 5]
    >>> y_pred = [1.1, 2.3, 0.4, 3.9, 3.1, 4.6]
    >>> max_error(y_true, y_pred)
    1.6

    """
    y_true, y_pred = _convert_to_ndarray(y_true, y_pred)
    _check_input(y_true, y_pred)

    error = np.amax(np.absolute(np.subtract(y_true, y_pred)))
    return error



from .base import TDAFeatures
from .amplitude_feature import AmplitudeFeature
from .average_lifetime_feature import AvgLifeTimeFeature
from .betti_curves_feature import BettiCurvesFeature
from .relevant_holes_feature import NumberOfRelevantHolesFeature

__all__ = [
    "TDAFeatures",
    "AmplitudeFeature",
    "AvgLifeTimeFeature",
    "BettiCurvesFeature",
    "NumberOfRelevantHolesFeature",
]



import numpy as np
import pandas as pd
import pytest

from giottotime.feature_creation.index_dependent_features.tda_features import (
    BettiCurvesFeature,
)


def test_correct_betti_curves_mean():
    np.random.seed(0)

    output_name = "betti_mean"
    betti_curve_feature = BettiCurvesFeature(output_name=output_name, betti_mode="mean")
    df = pd.DataFrame(np.random.randint(0, 100, size=(14, 1)), columns=["old_name"])

    betti_curves_mean = betti_curve_feature.fit_transform(df)

    expected_betti_curves = pd.DataFrame.from_dict(
        {
            f"{output_name}_0": [
                49.0,
                49.0,
                49.0,
                49.0,
                49.0,
                49.0,
                49.0,
                49.0,
                49.0,
                49.0,
                49.0,
                49.0,
                49.0,
                49.0,
            ],
            f"{output_name}_1": [
                59.325,
                59.325,
                59.325,
                59.325,
                59.325,
                59.325,
                59.325,
                45.640,
                45.640,
                45.640,
                45.640,
                45.640,
                45.640,
                45.640,
            ],
            f"{output_name}_2": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        }
    )

    assert np.allclose(expected_betti_curves, betti_curves_mean)


def test_correct_betti_curves_argmax():
    np.random.seed(0)

    output_name = "betti_mean"
    betti_curve_feature = BettiCurvesFeature(
        output_name=output_name, betti_mode="arg_max"
    )
    df = pd.DataFrame(np.random.randint(0, 100, size=(14, 1)), columns=["old_name"])

    betti_curves_arg_max = betti_curve_feature.fit_transform(df)

    expected_betti_curves = pd.DataFrame.from_dict(
        {
            f"{output_name}_0": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            f"{output_name}_1": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            f"{output_name}_2": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        }
    )

    assert np.allclose(expected_betti_curves, betti_curves_arg_max)


def test_wrong_inputs_for_betti_mode():
    df = pd.DataFrame(np.random.randint(0, 100, size=(14, 1)), columns=["old_name"])

    betti_mode_1 = "wrong_mode"
    betti_curve_feature_1 = BettiCurvesFeature(
        output_name="ignored", betti_mode=betti_mode_1
    )
    betti_mode_2 = ""
    betti_curve_feature_2 = BettiCurvesFeature(
        output_name="ignored", betti_mode=betti_mode_2
    )
    with pytest.raises(ValueError):
        betti_curve_feature_1.fit_transform(df)

    with pytest.raises(ValueError):
        betti_curve_feature_2.fit_transform(df)


def test_correct_mean_mode_betti_features():
    np.random.seed(0)

    df = pd.DataFrame(np.random.randint(0, 100, size=(14, 1)), columns=["old_name"])

    betti_mode_mean = "mean"

    betti_curves_mean = BettiCurvesFeature(
        output_name="ignored", betti_mode=betti_mode_mean
    )

    persistence_diagrams = betti_curves_mean._compute_persistence_diagrams(df)
    betti_curves = betti_curves_mean._compute_betti_curves(persistence_diagrams)

    betti_features = betti_curves_mean._compute_betti_features(betti_curves)
    expected_betti_features = [
        np.array([49.0, 49.0]),
        np.array([59.325, 45.64]),
        np.array([0.0, 0.0]),
    ]

    assert np.allclose(expected_betti_features, betti_features)


def test_correct_argmax_mode_betti_features():
    np.random.seed(0)

    df = pd.DataFrame(np.random.randint(0, 100, size=(14, 1)), columns=["old_name"])

    betti_mode_mean = "arg_max"

    betti_curves_mean = BettiCurvesFeature(
        output_name="ignored", betti_mode=betti_mode_mean
    )

    persistence_diagrams = betti_curves_mean._compute_persistence_diagrams(df)
    betti_curves = betti_curves_mean._compute_betti_curves(persistence_diagrams)

    betti_features = betti_curves_mean._compute_betti_features(betti_curves)
    expected_betti_features = [
        np.array([0.0, 0.0]),
        np.array([0.0, 0.0]),
        np.array([0.0, 0.0]),
    ]

    assert np.allclose(expected_betti_features, betti_features)



import numpy as np
import pandas as pd
import pytest
from hypothesis import given, strategies as st
from pandas.util import testing as testing

from giottotime.utils.hypothesis.time_indexes import giotto_time_series
from giottotime.feature_creation import (
    ConstantFeature,
    CustomFeature,
    PeriodicSeasonalFeature,
)


class TestPeriodicSesonalFeature:
    def test_missing_start_date_or_period(self):
        periodic_feature = PeriodicSeasonalFeature()
        with pytest.raises(ValueError):
            periodic_feature.transform()

        periodic_feature = PeriodicSeasonalFeature(index_period=1)
        with pytest.raises(ValueError):
            periodic_feature.transform()

        periodic_feature = PeriodicSeasonalFeature(start_date="2010-01-01")
        with pytest.raises(ValueError):
            periodic_feature.transform()

    def test_string_period(self):
        testing.N, testing.K = 20, 1
        ts = testing.makeTimeDataFrame(freq="s")
        periodic_feature = PeriodicSeasonalFeature(period="1 days")
        periodic_feature.transform(ts)

        assert type(periodic_feature.period) == pd.Timedelta

    def test_correct_start_date(self):
        testing.N, testing.K = 20, 1
        ts = testing.makeTimeDataFrame(freq="s")
        start_date = "2018-01-01"
        periodic_feature = PeriodicSeasonalFeature(
            period="1 days", start_date=start_date
        )
        periodic_feature.transform(ts)

        assert periodic_feature.start_date == ts.index.values[0]

        periodic_feature = PeriodicSeasonalFeature(
            period="3 days", index_period=10, start_date=start_date
        )
        periodic_feature.transform()
        assert periodic_feature.start_date == pd.to_datetime(start_date)

        start_date = pd.to_datetime("2018-01-01")
        periodic_feature = PeriodicSeasonalFeature(
            period="3 days", index_period=10, start_date=start_date
        )
        periodic_feature.transform()
        assert periodic_feature.start_date == start_date

    def test_too_high_sampling_frequency(self):
        start_date = "2018-01-01"
        periodic_feature = PeriodicSeasonalFeature(
            period="2 days",
            start_date=start_date,
            index_period=pd.DatetimeIndex(start=start_date, end="2020-01-01", freq="W"),
        )
        with pytest.raises(ValueError):
            periodic_feature.transform()

    def test_correct_sinusoide(self):
        testing.N, testing.K = 30, 1
        ts = testing.makeTimeDataFrame(freq="MS")
        start_date = "2018-01-01"
        periodic_feature = PeriodicSeasonalFeature(
            period="365 days",
            start_date=start_date,
            index_period=pd.DatetimeIndex(start=start_date, end="2020-01-01", freq="W"),
        )
        output_sin = periodic_feature.transform(ts)
        expected_index = pd.DatetimeIndex(
            [
                "2000-01-01",
                "2000-02-01",
                "2000-03-01",
                "2000-04-01",
                "2000-05-01",
                "2000-06-01",
                "2000-07-01",
                "2000-08-01",
                "2000-09-01",
                "2000-10-01",
                "2000-11-01",
                "2000-12-01",
                "2001-01-01",
                "2001-02-01",
                "2001-03-01",
                "2001-04-01",
                "2001-05-01",
                "2001-06-01",
                "2001-07-01",
                "2001-08-01",
                "2001-09-01",
                "2001-10-01",
                "2001-11-01",
                "2001-12-01",
                "2002-01-01",
                "2002-02-01",
                "2002-03-01",
                "2002-04-01",
                "2002-05-01",
                "2002-06-01",
            ],
            dtype="datetime64[ns]",
            freq="MS",
        )
        expected_df = pd.DataFrame.from_dict(
            {
                "PeriodicSeasonalFeature": [
                    0.0,
                    0.25433547,
                    0.42938198,
                    0.49999537,
                    0.43585316,
                    0.25062091,
                    0.0043035,
                    -0.25062091,
                    -0.43585316,
                    -0.49999537,
                    -0.42938198,
                    -0.24688778,
                    0.00860668,
                    0.2617078,
                    0.42938198,
                    0.49999537,
                    0.43585316,
                    0.25062091,
                    0.0043035,
                    -0.25062091,
                    -0.43585316,
                    -0.49999537,
                    -0.42938198,
                    -0.24688778,
                    0.00860668,
                    0.2617078,
                    0.42938198,
                    0.49999537,
                    0.43585316,
                    0.25062091,
                ]
            }
        )
        expected_df.index = expected_index
        pd.testing.assert_frame_equal(output_sin, expected_df)


class TestConstantFeature:
    def _correct_constant(
        self, df: pd.DataFrame, constant: int, output_name: str
    ) -> pd.DataFrame:
        constant_series = pd.Series(data=constant, index=df.index)
        constant_series.name = output_name
        return constant_series.to_frame()

    def test_correct_constant_feature(self):
        output_name = "constant_feature"
        constant = 12
        df = pd.DataFrame.from_dict({"old_name": [0, 1, 2, 3, 4, 5]})

        constant_feature = ConstantFeature(constant=constant, output_name=output_name)

        df_constant = constant_feature.fit_transform(df)
        expected_df_constant = pd.DataFrame.from_dict(
            {output_name: [constant, constant, constant, constant, constant, constant]}
        )

        testing.assert_frame_equal(expected_df_constant, df_constant)

    @given(
        giotto_time_series(
            min_length=1,
            start_date=pd.Timestamp(2000, 1, 1),
            end_date=pd.Timestamp(2010, 1, 1),
        ),
        st.integers(0, 100),
    )
    def test_random_ts_and_constant(self, df: pd.DataFrame, constant: int):
        output_name = "time_series"

        constant_feature = ConstantFeature(constant=constant, output_name=output_name)
        df_constant = constant_feature.fit_transform(df)
        expected_df_constant = self._correct_constant(df, constant, output_name)

        testing.assert_frame_equal(expected_df_constant, df_constant)


class TestCustomFeature:
    def _df_to_power(self, df: pd.DataFrame, power: int) -> pd.DataFrame:
        return np.power(df, power)

    def test_correct_custom_feature(self):
        output_name = "custom"
        power = 3
        df = pd.DataFrame.from_dict({"old_name": [0, 1, 2, 3, 4, 5]})

        custom_feature = CustomFeature(
            custom_feature_function=self._df_to_power,
            output_name=output_name,
            power=power,
        )

        output_custom_feature = custom_feature.fit_transform(df)
        expected_custom_output = pd.DataFrame.from_dict(
            {output_name: [0, 1, 8, 27, 64, 125]}
        )

        testing.assert_frame_equal(expected_custom_output, output_custom_feature)

    def test_multi_columns_custom_feature(self):
        output_name = "custom"
        power = 2
        df = pd.DataFrame.from_dict(
            {"old_name": [0, 1, 2, 3, 4, 5], "old_name_1": [7, 8, 9, 10, 11, 12]}
        )

        custom_feature = CustomFeature(
            custom_feature_function=self._df_to_power,
            output_name=output_name,
            power=power,
        )

        output_custom_feature = custom_feature.fit_transform(df)
        expected_custom_output = pd.DataFrame.from_dict(
            {
                f"{output_name}_0": [0, 1, 4, 9, 16, 25],
                f"{output_name}_1": [49, 64, 81, 100, 121, 144],
            }
        )

        testing.assert_frame_equal(expected_custom_output, output_custom_feature)

    @given(
        giotto_time_series(
            start_date=pd.Timestamp(2000, 1, 1), end_date=pd.Timestamp(2010, 1, 1)
        ),
        st.integers(0, 10),
    )
    def test_random_ts_and_power(self, df: pd.DataFrame, power: int):
        output_name = "time_series"

        custom_feature = CustomFeature(self._df_to_power, output_name, power=power)

        output_custom = custom_feature.fit_transform(df)
        expected_custom_output = self._df_to_power(df, power)

        testing.assert_frame_equal(expected_custom_output, output_custom)



from abc import ABC, abstractmethod
from typing import Optional, Union, List

import numpy as np
import pandas as pd

PandasTimeIndex = Union[pd.DatetimeIndex, pd.PeriodIndex, pd.TimedeltaIndex]
PandasDate = Union[pd.datetime, pd.Timestamp, str]

DEFAULT_START = pd.Timestamp("1970-01-01")
DEFAULT_END = pd.Timestamp("2020-01-01")
DEFAULT_FREQ = pd.Timedelta("1D")

__all__ = [
    "SequenceToTimeIndexSeries",
    "PandasSeriesToTimeIndexSeries",
    "TimeIndexSeriesToPeriodIndexSeries",
]


def count_not_none(*args):
    """Returns the count of arguments that are not None.
    """
    return sum(x is not None for x in args)


def check_period_range_parameters(
    start_date: PandasDate, end_date: PandasDate, periods: int
) -> None:
    """Check if the period range parameters given as input are compatible with the
    ``pd.period_range`` method.

    Of the three parameters: start, end, and periods, exactly two must be specified.

    Parameters
    ----------
    start_date : PandasDate, required
        The date to use as start date.

    end_date : PandasDate, required
        The date to use as end date.

    periods : int, required
        The number of periods.

    Raises
    ------
    ValueError
        Of the three parameters: start, end, and periods, exactly two must be specified.

    """
    if count_not_none(start_date, end_date, periods) != 2:
        raise ValueError(
            "Of the three parameters: start, end, and periods, "
            "exactly two must be specified"
        )


class TimeSeriesConversion(ABC):
    """Parent class for all time series type conversions.

    Subclasses must implement the two methods `_get_index_from` and `_get_values_from`.

    Parameters
    ----------
    start : PandasData, optional, default: ``None``
        start date of the output time series. Not mandatory for all time series
        conversion.

    end : PandasData, optional, default: ``None``
        end date of the output time series. Not mandatory for all time series
        conversion.

    freq : pd.Timedelta, optional, default: ``None``
        The frequency of the output time series. Not mandatory for all time series
        conversion.

    """

    def __init__(
        self,
        start: Optional[PandasDate] = None,
        end: Optional[PandasDate] = None,
        freq: Optional[pd.Timedelta] = None,
    ) -> None:
        self._initialize_start_end_freq(start, end, freq)

    def transform(self, time_series: Union[pd.Series, np.array, list]) -> pd.Series:
        """Transforms an array-like object (list, np.array, pd.Series) into a pd.Series
        with time index.

        It calls internally the abstract methods `_get_index_from()` and
        `_get_values_from()`. These are implemented in the subclasses.

        Parameters
        ----------
        time_series : Union[List, np.array, pd.Series], required
            It depends on the implementation of the subclasses.

        Returns
        -------
        time_series_t: pd.Series

        """
        index = self._get_index_from(time_series)
        values = self._get_values_from(time_series)
        return pd.Series(data=values, index=index)

    @abstractmethod
    def _get_index_from(
        self, array_like_object: Union[pd.Series, np.ndarray, list]
    ) -> PandasTimeIndex:
        raise NotImplementedError  # To exclude it from pytest coverage

    @abstractmethod
    def _get_values_from(
        self, array_like_object: Union[pd.Series, np.array, list]
    ) -> np.ndarray:
        raise NotImplementedError  # To exclude it from pytest coverage

    def _initialize_start_end_freq(
        self, start: PandasDate, end: PandasDate, freq: pd.Timedelta
    ) -> None:
        not_none_params = count_not_none(start, end, freq)
        if not_none_params == 0:
            self._default_params_initialization()
        elif not_none_params == 1:
            self._one_not_none_param_initialization(start, end, freq)
        elif not_none_params == 2:
            self._two_not_none_params_initialization(start, end, freq)
        else:
            raise ValueError(
                "Of the three parameters: start, end, and "
                "freq, exactly two must be specified"
            )

    def _default_params_initialization(self):
        self.start = DEFAULT_START
        self.end = None
        self.freq = DEFAULT_FREQ

    def _one_not_none_param_initialization(
        self, start: PandasDate, end: PandasDate, freq: pd.Timedelta
    ):
        if start is not None:
            self.start = start
            self.end = None
            self.freq = DEFAULT_FREQ
        elif end is not None:
            self.start = None
            self.end = end
            self.freq = DEFAULT_FREQ
        else:
            self.start = DEFAULT_START
            self.end = None
            self.freq = freq

    def _two_not_none_params_initialization(
        self, start: PandasDate, end: PandasDate, freq: pd.Timedelta
    ):
        self.start = start
        self.end = end
        self.freq = freq

    def _compute_period_index_of_length(self, length: int) -> pd.PeriodIndex:
        check_period_range_parameters(self.start, self.end, length)
        return pd.period_range(
            start=self.start, end=self.end, periods=length, freq=self.freq
        )


class SequenceToTimeIndexSeries(TimeSeriesConversion):
    """Converts a np.array or list to Series with PeriodIndex.

    Parameters
    -----------
    start : PandasDate, optional, default: ``None``
        start date of the output time series. Not mandatory for all time series
        conversion.

    end : PandasDate, optional, default: ``None``
        end date of the output time series. Not mandatory for all time series
        conversion.

    freq : pd.Timedelta, optional, default: ``None``
        frequency of the output time series. Not mandatory for all time series
        conversion.

    Examples
    --------
    >>> from giottotime.time_series_preparation import SequenceToTimeIndexSeries
    >>> time_series = [1,2,3,5,5,7]
    >>> sequence_to_time_index = SequenceToTimeIndexSeries(start='01-01-2010', freq='10D')
    >>> sequence_to_time_index.transform(time_series)
    2010-01-01    1
    2010-01-11    2
    2010-01-21    3
    2010-01-31    5
    2010-02-10    5
    2010-02-20    7
    Freq: 10D, dtype: int64
    """

    def __init__(
        self,
        start: Optional[PandasDate] = None,
        end: Optional[PandasDate] = None,
        freq: Optional[pd.Timedelta] = None,
    ) -> None:
        super().__init__(start, end, freq)

    def _get_index_from(
        self, array_like_object: Union[np.array, List[float]]
    ) -> pd.PeriodIndex:
        return self._compute_period_index_of_length(len(array_like_object))

    def _get_values_from(
        self, array_like_object: Union[np.array, List[float]]
    ) -> np.array:
        return np.array(array_like_object)


class PandasSeriesToTimeIndexSeries(TimeSeriesConversion):
    """Returns a Pandas Series with time index (DatetimeIndex, TimedeltaIndex or
    PeriodIndex) from a standard Pandas Series

    Parameters
    ----------
    start: PandasDate, required
        The date to use as start date.

    end: PandasDate, required
        The date to use as end date.

    freq : pd.Timedelta``, optional, default: ``None``
        The frequency of the time series.

    Examples
    --------
    >>> import pandas as pd
    >>> from giottotime.time_series_preparation import PandasSeriesToTimeIndexSeries
    >>> time_series = pd.Series([1,2,3,5,5,7])
    >>> sequence_to_time_index = PandasSeriesToTimeIndexSeries(start='01-01-2010', freq='10D')
    >>> sequence_to_time_index.transform(time_series)
    2010-01-01    1
    2010-01-11    2
    2010-01-21    3
    2010-01-31    5
    2010-02-10    5
    2010-02-20    7
    Freq: 10D, dtype: int64
    """

    def __init__(
        self,
        start: Optional[Union[pd.datetime, str]] = None,
        end: Optional[Union[pd.datetime, str]] = None,
        freq: Optional[pd.Timedelta] = None,
    ) -> None:
        super().__init__(start, end, freq)

    def _get_index_from(
        self, array_like_object: Union[pd.Series, np.array, list]
    ) -> PandasTimeIndex:
        if self._has_time_index(array_like_object):
            return array_like_object.index
        else:
            return self._compute_period_index_of_length(array_like_object.shape[0])

    def _get_values_from(
        self, array_like_object: Union[pd.Series, np.array, list]
    ) -> np.array:
        return array_like_object.values

    def _has_time_index(self, time_series: pd.Series) -> bool:
        index = time_series.index
        return (
            isinstance(index, pd.DatetimeIndex)
            or isinstance(index, pd.TimedeltaIndex)
            or isinstance(index, pd.PeriodIndex)
        )


class TimeIndexSeriesToPeriodIndexSeries(TimeSeriesConversion):
    """Converts a series with a time index (DatetimeIndex, TimedeltaIndex or
    PeriodIndex) to a series with a PeriodIndex.

    It may be necessary to specify a `freq` if not already provided.

    Parameters
    ----------
    freq : pd.Timedelta, optional, default: ``None``
        The frequency of the time series.

    Examples
    --------
    >>> import pandas as pd
    >>> from giottotime.time_series_preparation import TimeIndexSeriesToPeriodIndexSeries
    >>> period_index_time_series = pd.Series(
    ...     index = pd.period_range(start='01-01-2010', freq='10D', periods=6),
    ...     data=[1,2,3,5,5,7]
    ... )
    >>> datetime_index_time_series = pd.Series(
    ...     index = pd.date_range(start='01-01-2010', freq='10D', periods=6),
    ...     data=[1,2,3,5,5,7]
    ... )
    >>> timedelta_index_time_series = pd.Series(
    ...     index = pd.timedelta_range(start=pd.Timedelta(days=1), freq='10D', periods=6),
    ...     data=[1,2,3,5,5,7]
    ... )
    >>> sequence_to_time_index = TimeIndexSeriesToPeriodIndexSeries()
    >>> sequence_to_time_index.transform(period_index_time_series)
    2010-01-01    1
    2010-01-11    2
    2010-01-21    3
    2010-01-31    5
    2010-02-10    5
    2010-02-20    7
    freq: 10d, dtype: int64
    >>> sequence_to_time_index.transform(datetime_index_time_series)
    2010-01-01    1
    2010-01-11    2
    2010-01-21    3
    2010-01-31    5
    2010-02-10    5
    2010-02-20    7
    freq: 10d, dtype: int64
    >>> sequence_to_time_index.transform(timedelta_index_time_series)
    1970-01-02    1
    1970-01-12    2
    1970-01-22    3
    1970-02-01    5
    1970-02-11    5
    1970-02-21    7
    Freq: D, dtype: int64
    """

    def __init__(self, freq: Optional[pd.Timedelta] = None):
        super().__init__(start=None, end=None, freq=freq)

    def _get_index_from(self, time_series: pd.Series) -> pd.PeriodIndex:
        index = time_series.index
        if isinstance(index, pd.PeriodIndex):
            return index
        elif isinstance(index, pd.DatetimeIndex):
            return self._datetime_index_to_period(index)
        elif isinstance(index, pd.TimedeltaIndex):
            return self._timedelta_index_to_period(index)
        else:
            raise ValueError(
                f"Only PeriodIndex, DatetimeIndex and "
                f"TimedeltaIndex are supported. Detected: "
                f"{type(index)}"
            )

    def _datetime_index_to_period(self, index: pd.DatetimeIndex) -> pd.PeriodIndex:
        if index.freq is None:
            return pd.PeriodIndex(index, freq=self.freq)
        else:
            return pd.PeriodIndex(index)

    def _timedelta_index_to_period(self, index: pd.TimedeltaIndex) -> pd.PeriodIndex:
        datetime_index = pd.to_datetime(index)
        return self._datetime_index_to_period(datetime_index)

    def _get_values_from(self, time_series: pd.Series) -> np.array:
        return time_series.values



import numpy as np
import pandas as pd
import pytest
from hypothesis import given
from hypothesis.extra.numpy import arrays
from hypothesis.strategies import floats


from giottotime.loss_functions import smape, max_error


class TestSmape:
    def _correct_smape(self, y_true, y_pred):
        non_normalized_smape = sum(
            np.abs(y_pred - y_true) / (np.abs(y_pred) + np.abs(y_true))
        )
        non_normalized_smape_filled = np.nan_to_num(non_normalized_smape, nan=0)
        error = (2 / len(y_pred)) * non_normalized_smape_filled
        return error

    def test_wrong_vector_length(self):
        y_true = np.random.random(5)
        y_pred = np.random.random(4)

        with pytest.raises(ValueError):
            smape(y_true, y_pred)

    def test_nan_values(self):
        y_true = np.array([np.nan, 1, 2, 3])
        y_pred = np.random.random(4)

        with pytest.raises(ValueError):
            smape(y_true, y_pred)

    def test_infinite_values(self):
        y_true = np.random.random(4)
        y_pred = np.array([0, np.inf, 2, 3])

        with pytest.raises(ValueError):
            smape(y_true, y_pred)

    def test_correct_smape_array(self):
        y_true = np.array([0, 1, 2, 3, 4, 5])
        y_pred = np.array([3, 2, 4, 3, 5, 7])

        error = smape(y_true, y_pred)
        expected_error = (1 + 1 / 3 + 1 / 3 + 0 + 1 / 9 + 1 / 6) * (2 / 6)

        assert expected_error == error

    def test_correct_smape_dataframe(self):
        y_true = pd.DataFrame([0, 1, 2, 3, 4, 5])
        y_pred = pd.DataFrame([3, 2, 4, 3, 5, 7])

        error = smape(y_true, y_pred)
        expected_error = (1 + 1 / 3 + 1 / 3 + 0 + 1 / 9 + 1 / 6) * (2 / 6)

        assert expected_error == error

    def test_correct_smape_list(self):
        y_true = [0, 1, 2, 3, 4, 5]
        y_pred = [3, 2, 4, 3, 5, 7]

        error = smape(y_true, y_pred)
        expected_error = (1 + 1 / 3 + 1 / 3 + 0 + 1 / 9 + 1 / 6) * (2 / 6)

        assert expected_error == error

    def test_smape_is_symmetric(self):
        y_true = np.random.random(10)
        y_pred = np.random.random(10)

        first_error = smape(y_true, y_pred)
        second_error = smape(y_pred, y_true)

        assert first_error == second_error

    @given(
        arrays(float, shape=30, elements=floats(allow_nan=False, allow_infinity=False)),
        arrays(float, shape=30, elements=floats(allow_nan=False, allow_infinity=False)),
    )
    def test_between_zero_and_one(self, y_true, y_pred):
        error = smape(y_true, y_pred)

        assert error >= 0
        assert error <= 2

    @given(
        arrays(float, shape=30, elements=floats(allow_nan=False, allow_infinity=False)),
        arrays(float, shape=30, elements=floats(allow_nan=False, allow_infinity=False)),
    )
    def test_smape_random_arrays(self, y_true, y_pred):
        error = smape(y_true, y_pred)
        expected_error = self._correct_smape(y_true, y_pred)

        assert expected_error == error


class TestMaxError:
    def _correct_max_error(self, y_true, y_pred):
        error = np.amax(np.absolute(np.subtract(y_true, y_pred)))
        return error

    def test_wrong_vector_length(self):
        y_true = np.random.random(5)
        y_pred = np.random.random(4)

        with pytest.raises(ValueError):
            max_error(y_true, y_pred)

    def test_nan_values(self):
        y_true = [np.nan, 1, 2, 3]
        y_pred = np.random.random(4)

        with pytest.raises(ValueError):
            max_error(y_true, y_pred)

    def test_infinite_values(self):
        y_true = np.random.random(4)
        y_pred = [0, np.inf, 2, 3]

        with pytest.raises(ValueError):
            max_error(y_true, y_pred)

    def test_max_error_list(self):
        y_true = [0, 1, 2, 3, 4, 5]
        y_pred = [-1, 4, 5, 10, 4, 1]

        error = max_error(y_pred, y_true)
        expected_max_error = 7

        assert expected_max_error, error

    def test_max_error_array(self):
        y_true = np.array([0, 1, 2, 3, 4, 5])
        y_pred = np.array([-1, 4, 5, 10, 4, 1])

        error = max_error(y_pred, y_true)
        expected_max_error = 7

        assert expected_max_error, error

    def test_max_error_dataframe(self):
        y_true = pd.DataFrame([0, 1, 2, 3, 4, 5])
        y_pred = pd.DataFrame([-1, 4, 5, 10, 4, 1])

        error = max_error(y_pred, y_true)
        expected_max_error = 7

        assert expected_max_error, error

    @given(
        arrays(float, shape=30, elements=floats(allow_nan=False, allow_infinity=False)),
        arrays(float, shape=30, elements=floats(allow_nan=False, allow_infinity=False)),
    )
    def test_smape_random_arrays_finite_values(self, y_true, y_pred):
        error = max_error(y_true, y_pred)
        expected_error = self._correct_max_error(y_true, y_pred)
        print(y_true)
        print(y_pred)
        assert expected_error == error



from .base import IndexDependentFeature
from .calendar_features import CalendarFeature
from .time_series_features import (
    ShiftFeature,
    MovingAverageFeature,
    PolynomialFeature,
    ExogenousFeature,
)
from .trend_features import (
    DetrendedFeature,
    RemovePolynomialTrend,
    RemoveExponentialTrend,
)
from .tda_features import (
    AmplitudeFeature,
    AvgLifeTimeFeature,
    BettiCurvesFeature,
    NumberOfRelevantHolesFeature,
)

__all__ = [
    "IndexDependentFeature",
    "CalendarFeature",
    "ShiftFeature",
    "MovingAverageFeature",
    "PolynomialFeature",
    "ExogenousFeature",
    "DetrendedFeature",
    "RemovePolynomialTrend",
    "RemoveExponentialTrend",
    "AmplitudeFeature",
    "AvgLifeTimeFeature",
    "BettiCurvesFeature",
    "NumberOfRelevantHolesFeature",
]



#! /usr/bin/env python
"""Toolbox for Machine Learning using Topological Data Analysis."""

import os
import codecs

from setuptools import setup, find_packages

from giottotime import __version__

version_file = os.path.join("giottotime", "_version.py")
with open(version_file) as f:
    exec(f.read())

with open("requirements.txt") as f:
    requirements = f.read().splitlines()
with open("doc-requirements.txt") as f:
    doc_requirements = f.read().splitlines()
with open("dev-requirements.txt") as f:
    dev_requirements = f.read().splitlines()

DISTNAME = "giotto-time"
DESCRIPTION = "Toolbox for Time Series analysis and integration with Machine Learning."
with codecs.open("README.rst", encoding="utf-8-sig") as f:
    LONG_DESCRIPTION = f.read()
LONG_DESCRIPTION_TYPE = "text/x-rst"
MAINTAINER = "Alessio Baccelli"
MAINTAINER_EMAIL = "maintainers@giotto.ai"
URL = "https://github.com/giotto-ai/giotto-time"
LICENSE = "GPLv3"
DOWNLOAD_URL = "https://github.com/giotto-ai/giotto-time/tarball/v0.0a0"
VERSION = __version__
CLASSIFIERS = [
    "Intended Audience :: Information Technology",
    "Intended Audience :: Developers",
    "License :: OSI Approved",
    "Programming Language :: Python",
    "Topic :: Software Development",
    "Topic :: Scientific/Engineering",
    "Operating System :: Microsoft :: Windows",
    "Operating System :: POSIX",
    "Operating System :: Unix",
    "Operating System :: MacOS",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
]
KEYWORDS = (
    "machine learning time series data analysis " + "topology, persistence diagrams"
)
INSTALL_REQUIRES = requirements
EXTRAS_REQUIRE = {
    "tests": dev_requirements,
    "doc": doc_requirements,
    "examples": [],
}


setup(
    name=DISTNAME,
    maintainer=MAINTAINER,
    maintainer_email=MAINTAINER_EMAIL,
    description=DESCRIPTION,
    license=LICENSE,
    url=URL,
    version=VERSION,
    download_url=DOWNLOAD_URL,
    long_description=LONG_DESCRIPTION,
    long_description_content_type=LONG_DESCRIPTION_TYPE,
    zip_safe=False,
    classifiers=CLASSIFIERS,
    packages=find_packages(),
    keywords=KEYWORDS,
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
)



import numpy as np
import pandas as pd
import pandas.util.testing as testing
import pytest
from hypothesis import given, strategies as st
from sklearn.preprocessing import PolynomialFeatures

from giottotime.utils.hypothesis.time_indexes import giotto_time_series
from giottotime.feature_creation import (
    ShiftFeature,
    MovingAverageFeature,
    ExogenousFeature,
    PolynomialFeature,
)


class TestShiftFeature:
    def _correct_shift(self, df: pd.DataFrame, shift: int) -> pd.DataFrame:
        return df.shift(shift)

    def test_forward_time_shift_feature(self):
        output_name = "shift"
        df = pd.DataFrame.from_dict({"old_name": [0, 1, 2, 3, 4, 5]})

        shift_feature = ShiftFeature(shift=1, output_name=output_name)
        expected_df = pd.DataFrame.from_dict({output_name: [np.nan, 0, 1, 2, 3, 4]})

        df_shifted = shift_feature.fit_transform(df)
        testing.assert_frame_equal(expected_df, df_shifted)

    def test_backwards_time_shift_feature(self):
        output_name = "shift"
        df = pd.DataFrame.from_dict({"old_name": [0, 1, 2, 3, 4, 5]})

        shift_feature = ShiftFeature(shift=-2, output_name=output_name)
        expected_df = pd.DataFrame.from_dict(
            {output_name: [2, 3, 4, 5, np.nan, np.nan]}
        )

        df_shifted = shift_feature.fit_transform(df)
        testing.assert_frame_equal(expected_df, df_shifted)

    def test_zero_shift_feature(self):
        output_name = "shift"
        df = pd.DataFrame.from_dict({"old_name": [0, 1, 2, 3, 4, 5]})

        shift_feature = ShiftFeature(shift=0, output_name=output_name)
        expected_df = pd.DataFrame.from_dict({output_name: [0, 1, 2, 3, 4, 5]})

        df_shifted = shift_feature.fit_transform(df)
        testing.assert_frame_equal(expected_df, df_shifted)

    def test_multi_columns_time_shift_feature(self):
        output_name = "shift"
        df = pd.DataFrame.from_dict(
            {"old_name_0": [0, 1, 2, 3, 4, 5], "old_name_1": [7, 8, 9, 10, 11, 12]}
        )

        shift_feature = ShiftFeature(shift=-2, output_name=output_name)
        expected_df = pd.DataFrame.from_dict(
            {
                f"{output_name}_0": [2, 3, 4, 5, np.nan, np.nan],
                f"{output_name}_1": [9, 10, 11, 12, np.nan, np.nan],
            }
        )

        df_shifted = shift_feature.fit_transform(df)
        testing.assert_frame_equal(expected_df, df_shifted)

    @given(
        giotto_time_series(
            start_date=pd.Timestamp(2000, 1, 1), end_date=pd.Timestamp(2010, 1, 1)
        ),
        st.integers(0, 200),
    )
    def test_random_ts_and_shifts(self, df: pd.DataFrame, shift: int):
        output_name = "time_series"
        shift_feature = ShiftFeature(shift=shift, output_name=output_name)

        df_shifted = shift_feature.fit_transform(df)
        correct_df_shifted = self._correct_shift(df, shift)

        testing.assert_frame_equal(correct_df_shifted, df_shifted)


class TestMovingAverageFeature:
    def _correct_ma(self, df: pd.DataFrame, window_size: int) -> pd.DataFrame:
        return df.rolling(window_size).mean()

    def test_invalid_window_size(self):
        output_name = "moving_average"
        window_size = -1
        df = pd.DataFrame.from_dict({"old_name": [0, 1, 2, 3, 4, 5]})

        ma_feature = MovingAverageFeature(
            window_size=window_size, output_name=output_name
        )
        with pytest.raises(ValueError):
            ma_feature.fit_transform(df)

    def test_positive_window_size(self):
        output_name = "moving_average"
        window_size = 2
        df = pd.DataFrame.from_dict({"old_name": [0, 1, 2, 3, 4, 5]})

        ma_feature = MovingAverageFeature(
            window_size=window_size, output_name=output_name
        )
        df_ma = ma_feature.fit_transform(df)
        expected_df_ma = pd.DataFrame.from_dict(
            {output_name: [np.nan, 0.5, 1.5, 2.5, 3.5, 4.5]}
        )

        testing.assert_frame_equal(expected_df_ma, df_ma)

    def test_multi_columns_window_size(self):
        output_name = "moving_average"
        window_size = 2
        df = pd.DataFrame.from_dict(
            {"old_name_0": [0, 1, 2, 3, 4, 5], "old_name_1": [7, 8, 9, 10, 11, 12]}
        )

        ma_feature = MovingAverageFeature(
            window_size=window_size, output_name=output_name
        )
        df_ma = ma_feature.fit_transform(df)
        expected_df_ma = pd.DataFrame.from_dict(
            {
                f"{output_name}_0": [np.nan, 0.5, 1.5, 2.5, 3.5, 4.5],
                f"{output_name}_1": [np.nan, 7.5, 8.5, 9.5, 10.5, 11.5],
            }
        )

        testing.assert_frame_equal(expected_df_ma, df_ma)

    @given(
        giotto_time_series(
            start_date=pd.Timestamp(2000, 1, 1), end_date=pd.Timestamp(2010, 1, 1)
        ),
        st.integers(0, 100),
    )
    def test_random_ts_and_window_size(self, df: pd.DataFrame, window_size: int):
        output_name = "time_series"

        ma_feature = MovingAverageFeature(
            window_size=window_size, output_name=output_name
        )
        df_ma = ma_feature.fit_transform(df)
        expected_df_ma = self._correct_ma(df, window_size)

        testing.assert_frame_equal(expected_df_ma, df_ma)


class TestExogenousFeature:
    def _correct_exog(self, exog: pd.DataFrame, df: pd.DataFrame) -> pd.DataFrame:
        return exog.reindex(index=df.index)

    def test_correct_exog_none_method(self):
        output_name = "exog"
        method = None
        exog = pd.DataFrame.from_dict({"old_name_1": [0, 1, 2, 3]})
        exog.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 2, 1),
            pd.Timestamp(2000, 3, 1),
            pd.Timestamp(2000, 4, 1),
        ]

        df = pd.DataFrame.from_dict({"old_name_2": [10, 11, 12, 13]})
        df.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 1, 2),
            pd.Timestamp(2000, 1, 3),
            pd.Timestamp(2000, 1, 4),
        ]

        exog_feature = ExogenousFeature(
            exogenous_time_series=exog, output_name=output_name, method=method
        )

        new_exog_feature = exog_feature.fit_transform(df)
        expected_exog = pd.DataFrame.from_dict(
            {output_name: [0, np.nan, np.nan, np.nan]}
        )
        expected_exog.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 1, 2),
            pd.Timestamp(2000, 1, 3),
            pd.Timestamp(2000, 1, 4),
        ]

        testing.assert_frame_equal(expected_exog, new_exog_feature)

    def test_correct_exog_backfill_method(self):
        output_name = "exog"
        method = "backfill"
        exog = pd.DataFrame.from_dict({"old_name_1": [0, 1, 2, 3]})
        exog.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 2, 1),
            pd.Timestamp(2000, 3, 1),
            pd.Timestamp(2000, 4, 1),
        ]

        df = pd.DataFrame.from_dict({"old_name_2": [10, 11, 12, 13]})
        df.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 1, 2),
            pd.Timestamp(2000, 1, 3),
            pd.Timestamp(2000, 1, 4),
        ]

        exog_feature = ExogenousFeature(
            exogenous_time_series=exog, output_name=output_name, method=method
        )

        new_exog_feature = exog_feature.fit_transform(df)
        expected_exog = pd.DataFrame.from_dict({output_name: [0, 1, 1, 1]})
        expected_exog.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 1, 2),
            pd.Timestamp(2000, 1, 3),
            pd.Timestamp(2000, 1, 4),
        ]

        testing.assert_frame_equal(expected_exog, new_exog_feature)

    def test_correct_exog_pad_method(self):
        output_name = "exog"
        method = "pad"
        exog = pd.DataFrame.from_dict({"old_name_1": [0, 1, 2, 3]})
        exog.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 2, 1),
            pd.Timestamp(2000, 3, 1),
            pd.Timestamp(2000, 4, 1),
        ]

        df = pd.DataFrame.from_dict({"old_name_2": [10, 11, 12, 13]})
        df.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 1, 2),
            pd.Timestamp(2000, 1, 3),
            pd.Timestamp(2000, 1, 4),
        ]

        exog_feature = ExogenousFeature(
            exogenous_time_series=exog, output_name=output_name, method=method
        )

        new_exog_feature = exog_feature.fit_transform(df)
        expected_exog = pd.DataFrame.from_dict({output_name: [0, 0, 0, 0]})
        expected_exog.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 1, 2),
            pd.Timestamp(2000, 1, 3),
            pd.Timestamp(2000, 1, 4),
        ]

        testing.assert_frame_equal(expected_exog, new_exog_feature)

    def test_correct_nearest_pad_method(self):
        output_name = "exog"
        method = "nearest"
        exog = pd.DataFrame.from_dict({"old_name_1": [0, 1, 2, 3]})
        exog.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 2, 1),
            pd.Timestamp(2000, 3, 1),
            pd.Timestamp(2000, 4, 1),
        ]

        df = pd.DataFrame.from_dict({"old_name_2": [10, 11, 12, 13]})
        df.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 1, 2),
            pd.Timestamp(2000, 1, 3),
            pd.Timestamp(2000, 1, 29),
        ]

        exog_feature = ExogenousFeature(
            exogenous_time_series=exog, output_name=output_name, method=method
        )

        new_exog_feature = exog_feature.fit_transform(df)
        expected_exog = pd.DataFrame.from_dict({output_name: [0, 0, 0, 1]})
        expected_exog.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 1, 2),
            pd.Timestamp(2000, 1, 3),
            pd.Timestamp(2000, 1, 29),
        ]

        testing.assert_frame_equal(expected_exog, new_exog_feature)

    def test_correct_multi_columns_exog(self):
        output_name = "exog"
        exog = pd.DataFrame.from_dict(
            {"old_name_0": [0, 1, 2, 3], "old_name_1": [5, 6, 7, 8]}
        )
        exog.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 2, 1),
            pd.Timestamp(2000, 3, 1),
            pd.Timestamp(2000, 4, 1),
        ]

        df = pd.DataFrame.from_dict({"old_name_2": [10, 11, 12, 13]})
        df.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 1, 2),
            pd.Timestamp(2000, 1, 3),
            pd.Timestamp(2000, 1, 29),
        ]

        exog_feature = ExogenousFeature(
            exogenous_time_series=exog, output_name=output_name
        )

        new_exog_feature = exog_feature.fit_transform(df)
        expected_exog = pd.DataFrame.from_dict(
            {
                f"{output_name}_0": [0, np.nan, np.nan, np.nan],
                f"{output_name}_1": [5, np.nan, np.nan, np.nan],
            }
        )
        expected_exog.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 1, 2),
            pd.Timestamp(2000, 1, 3),
            pd.Timestamp(2000, 1, 29),
        ]

        testing.assert_frame_equal(expected_exog, new_exog_feature)


class TestPolynomialFeature:
    def _correct_pol_features(self, df: pd.DataFrame, degree: int) -> pd.DataFrame:
        poly = PolynomialFeatures(degree)
        pol_features_arr = poly.fit_transform(df)
        pol_features = pd.DataFrame(pol_features_arr, index=df.index)
        return pol_features

    def test_correct_pol_features_single_column(self):
        output_name = "pol_features"
        degree = 3
        df = pd.DataFrame.from_dict({"old_name": [0, 1, 2, 3]})
        df.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 2, 1),
            pd.Timestamp(2000, 3, 1),
            pd.Timestamp(2000, 4, 1),
        ]

        pol_feature = PolynomialFeature(degree=degree, output_name=output_name)
        pol_df = pol_feature.fit_transform(df)

        expected_pol_df = pd.DataFrame.from_dict(
            {
                f"{output_name}_0": [1.0, 1.0, 1.0, 1.0],
                f"{output_name}_1": [0.0, 1.0, 2.0, 3.0],
                f"{output_name}_2": [0.0, 1.0, 4.0, 9.0],
                f"{output_name}_3": [0.0, 1.0, 8.0, 27.0],
            }
        )
        expected_pol_df.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 2, 1),
            pd.Timestamp(2000, 3, 1),
            pd.Timestamp(2000, 4, 1),
        ]

        testing.assert_frame_equal(expected_pol_df, pol_df)

    def test_correct_pol_features_multi_columns(self):
        output_name = "pol_features"
        degree = 2
        df = pd.DataFrame.from_dict({"old_name_1": [0, 2, 4], "old_name_2": [1, 3, 5]})
        df.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 2, 1),
            pd.Timestamp(2000, 3, 1),
        ]

        pol_feature = PolynomialFeature(degree=degree, output_name=output_name)
        pol_df = pol_feature.fit_transform(df)

        expected_pol_df = pd.DataFrame.from_dict(
            {
                f"{output_name}_0": [1.0, 1.0, 1],
                f"{output_name}_1": [0.0, 2.0, 4],
                f"{output_name}_2": [1.0, 3.0, 5],
                f"{output_name}_3": [0.0, 4.0, 16.0],
                f"{output_name}_4": [0.0, 6.0, 20.0],
                f"{output_name}_5": [1.0, 9.0, 25.0],
            }
        )
        expected_pol_df.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 2, 1),
            pd.Timestamp(2000, 3, 1),
        ]

        testing.assert_frame_equal(expected_pol_df, pol_df)



"""
The :mod:`giottotime.feature_creation` module deals with the creation of features
starting from a time series.
"""

from .feature_creation import FeatureCreation
from .index_dependent_features import (
    CalendarFeature,
    DetrendedFeature,
    RemovePolynomialTrend,
    RemoveExponentialTrend,
    tda_features,
)
from .index_dependent_features import (
    ShiftFeature,
    MovingAverageFeature,
    PolynomialFeature,
    ExogenousFeature,
    AmplitudeFeature,
    AvgLifeTimeFeature,
    BettiCurvesFeature,
    NumberOfRelevantHolesFeature,
)

from .standard_features import ConstantFeature, PeriodicSeasonalFeature, CustomFeature

__all__ = [
    "FeatureCreation",
    "ShiftFeature",
    "MovingAverageFeature",
    "ConstantFeature",
    "PolynomialFeature",
    "ExogenousFeature",
    "tda_features",
    "CalendarFeature",
    "PeriodicSeasonalFeature",
    "DetrendedFeature",
    "RemovePolynomialTrend",
    "RemoveExponentialTrend",
    "AmplitudeFeature",
    "AvgLifeTimeFeature",
    "BettiCurvesFeature",
    "NumberOfRelevantHolesFeature",
]



from typing import Union, Optional, Callable

import numpy as np
import pandas as pd
from pandas import DatetimeIndex

from math import pi

from giottotime.feature_creation.standard_features.base import StandardFeature

__all__ = ["PeriodicSeasonalFeature", "ConstantFeature", "CustomFeature"]


class PeriodicSeasonalFeature(StandardFeature):
    """Create a sinusoid from a given date and with a given period and
    amplitude.

    Parameters
    ----------
    period : Union[pd.Timedelta, str], optional, default: ``'365 days'``
        The period of the generated time series.

    amplitude : float, optional, default: ``0.5``
        The amplitude of the time series.

    start_date : Union[pd.Timestamp, str], optional, default: ``None``
        The date from which to start generating the feature. This is used only if X is
        not passed in the ``transform`` method, otherwise the start date is inferred
        from it.

    length : int, optional, default: ``50``
        The length of the sinusoid. This is used only if X is not passed in the
        ``transform`` method, otherwise the length is inferred from it.

    index_period : Union[DatetimeIndex, int], optional, default: ``None``
        The period of the index of the output ``DataFrame``. This is used only if X is
        not passed in the ``transform`` method, otherwise the index period is taken
        from it.

    output_name : str, optional, default: ``"PeriodicSeasonalFeature"``
        The name of the output column.

    Examples
    --------
    >>> from giottotime.feature_creation import PeriodicSeasonalFeature
    >>> period_feature = PeriodicSeasonalFeature(start_date="2020-01-01",
    ...                                          index_period=10)
    >>> period_feature.transform()
    Float64Index([              0.0, 0.0027397260273972603,
               0.005479452054794521,   0.00821917808219178,
               0.010958904109589041,    0.0136986301369863,
                0.01643835616438356,  0.019178082191780823,
               0.021917808219178082,  0.024657534246575342],
             dtype='float64')

       PeriodicSeasonalFeature
    0                 0.000000
    1                 0.008607
    2                 0.017211
    3                 0.025810
    4                 0.034401
    5                 0.042982
    6                 0.051551
    7                 0.060104
    8                 0.068639
    9                 0.077154

    """

    def __init__(
        self,
        period: Union[pd.Timedelta, str] = "365 days",
        amplitude: float = 0.5,
        start_date: Optional[Union[pd.Timestamp, str]] = None,
        length: Optional[int] = 50,
        index_period: Optional[Union[DatetimeIndex, int]] = None,
        output_name: str = "PeriodicSeasonalFeature",
    ):
        super().__init__(output_name=output_name)
        self.start_date = start_date
        self.period = period
        self.amplitude = amplitude
        self.length = length
        self.index_period = index_period

    def transform(self, time_series: Optional[pd.DataFrame] = None) -> pd.DataFrame:
        """Generate a sinusoid, with the given ``period``, ``amplitude`` and ``length``,
        starting from the selected ``start_date``. If ``time_series`` is not ``None``,
        the ``start_date`` is replaced by the start date of the time series and the
        output sinusoid will have the same index as ``time_series``.

        Parameters
        ----------
        time_series : pd.DataFrame, shape (n_samples, 1), optional, default: ``None``
            The input DataFrame, If passed, the output DataFrame is going to have the
            same index as ``time_series``. If is not passed, then the ``start_date`` and
             ``index_period`` must have been passed in the constructor when the object
             was instantiated.

        Returns
        -------
        periodic_feature : pd.DataFrame, shape (n_samples, 1)
            The DataFrame containing the generated period feature.

        Raises
        ------
        ValueError
            Raised if ``time_series`` is not passed and the ``start_date`` or the
            ``index_period`` are not present.

        """
        self._validate_input(time_series)
        self._transform_inputs(time_series)
        datetime_index = self._get_time_index(time_series)

        periodic_feature_values = self._compute_periodic_feature(datetime_index)
        if time_series is not None:
            periodic_feature = pd.DataFrame(
                index=time_series.index, data=periodic_feature_values
            )

        else:
            periodic_feature = pd.DataFrame(data=periodic_feature_values[: self.length])

        periodic_feature = self._rename_columns(periodic_feature)
        return periodic_feature

    def _validate_input(self, X: pd.DataFrame) -> None:
        if X is None:
            if self.start_date is None:
                raise ValueError(
                    f"If X is not provided, the 'start_date' variable "
                    f"must be provided, but instead was "
                    f"{self.start_date}."
                )
            if self.index_period is None:
                raise ValueError(
                    f"If X is not provided, the 'index_period' variable "
                    f"must be provided, but instead was "
                    f"{self.index_period}."
                )

    def _transform_inputs(self, X: pd.DataFrame) -> None:
        if isinstance(self.period, str):
            self.period = pd.to_timedelta(self.period)

        if X is not None:
            self.start_date = X.index.values[0]
        else:
            if isinstance(self.start_date, str):
                self.start_date = pd.to_datetime(self.start_date)

    def _get_time_index(self, X: pd.DataFrame) -> pd.DatetimeIndex:
        if X is not None:
            datetime_index = X.index
        else:
            if isinstance(self.index_period, int):
                datetime_index = pd.date_range(
                    start=self.start_date, periods=self.index_period
                )
            else:
                datetime_index = self.index_period

        self._check_sampling_frequency(datetime_index)

        return datetime_index

    def _check_sampling_frequency(self, datetime_index: pd.DatetimeIndex) -> None:
        if isinstance(datetime_index, pd.PeriodIndex):
            datetime_index = datetime_index.to_timestamp()
        sampling_frequency = pd.Timedelta(datetime_index[1] - datetime_index[0])

        if self.period < 2 * sampling_frequency:
            raise ValueError(
                f"Sampling frequency must be at least two times"
                f"the period to obtain meaningful results. "
                f"Sampling frequency = {sampling_frequency},"
                f"period = {self.period}."
            )

    def _compute_periodic_feature(self, datetime_index: pd.DatetimeIndex):
        print((datetime_index - self.start_date) / self.period)
        return (
            np.sin(2 * pi * (datetime_index - self.start_date) / self.period)
        ) * self.amplitude


class ConstantFeature(StandardFeature):
    """Generate a ``pd.DataFrame`` with one column, of the same length as the input
     ``X`` and containing the value ``constant`` across the whole column.

    Parameters
    ----------
    constant : int, optional, default: ``2``
        The value to use to generate the constant column of the ``pd.DataFrame``.

    length : int, optional, default: ``50``
        The length of the DataFrame to generate. This is used only if X is not passed in
        the ``transform`` method, otherwise the length is inferred from it.

    output_name : str, optional, default: ``'ConstantFeature'``
        The name of the output column.

    Examples
    --------
    >>> from giottotime.feature_creation import ConstantFeature
    >>> constant_feature = ConstantFeature(constant=3, length=5)
    >>> constant_feature.transform()
      ConstantFeature
    0               3
    1               3
    2               3
    3               3
    4               3

    """

    def __init__(
        self, constant: int = 2, length: int = 50, output_name: str = "ConstantFeature"
    ):
        super().__init__(output_name)
        self.length = length
        self.constant = constant

    def transform(self, time_series: Optional[pd.DataFrame] = None) -> pd.DataFrame:
        """Generate a ``pd.DataFrame`` with one column with the same length as
        ``time_series`` and with the same index, containing a value equal to
        ``constant``.

        Parameters
        ----------
        time_series : pd.DataFrame, shape (n_samples, 1), optional, default: ``None``
            The input DataFrame. If passed, the output DataFrame is going to have the
            same index as ``time_series``.

        Returns
        -------
        constant_series_renamed : pd.DataFrame, shape (length, 1)
            A constant series, with the same length of ``X`` and with the same index.

        """
        if time_series is not None:
            constant_series = pd.Series(
                data=self.constant, index=time_series.index
            ).to_frame()
        else:
            constant_series = pd.Series(data=[self.constant] * self.length).to_frame()

        constant_series_renamed = self._rename_columns(constant_series)
        return constant_series_renamed


class CustomFeature(StandardFeature):
    """Given a custom function, apply it to a time series and generate a
    ``pd.Dataframe``.

    Parameters
    ----------
    custom_feature_function : Callable, required.
        The function to use to generate a ``pd.DataFrame`` containing the feature.

    output_name: str, optional, default: ``'CustomFeature'``.
        The name of the output column.

    kwargs : ``object``, optional.
        Optional arguments to pass to the function.

    Examples
    --------
    >>> import pandas as pd
    >>> from giottotime.feature_creation import CustomFeature
    >>> def custom_function(X, power):
    ...     return X**power
    >>> X = pd.DataFrame([0, 1, 2, 3, 4, 5])
    >>> custom_feature = CustomFeature(custom_function, output_name="custom_f", power=3)
    >>> custom_feature.transform(X)
       custom_f
    0         0
    1         1
    2         8
    3        27
    4        64
    5       125

    """

    def __init__(
        self,
        custom_feature_function: Callable,
        output_name: str = "CustomFeature",
        **kwargs: object,
    ):
        super().__init__(output_name)
        self.custom_feature_function = custom_feature_function
        self.kwargs = kwargs

    def transform(self, time_series: Optional[pd.DataFrame] = None) -> pd.DataFrame:
        """Generate a ``pd.DataFrame``, given ``time_series`` as input to the
        ``custom_feature_function``, as well as other optional arguments.

        Parameters
        ----------
        time_series : pd.DataFrame, shape (n_samples, 1), optional, default: ``None``
            The DataFrame on which to apply the the custom function.

        Returns
        -------
        custom_feature_renamed : pd.DataFrame, shape (length, 1)
            A DataFrame containing the generated feature_creation.

        Notes
        -----
        In order to use the ``CustomFeature`` class inside a
        ``giottotime.feature_creation.FeatureCreation`` class, the output of  the custom
         function should be a ``pd.DataFrame`` and have the same index as
         ``time_series``.

        """
        custom_feature = self.custom_feature_function(time_series, **self.kwargs)
        custom_feature_renamed = self._rename_columns(custom_feature)
        return custom_feature_renamed



from abc import abstractmethod
from typing import Optional

import pandas as pd

from giottotime.feature_creation.base import Feature


class StandardFeature(Feature):
    """Base class for all the feature classes in this package.

    """

    def fit(self, X, y=None):
        return self

    @abstractmethod
    def transform(self, X: Optional[pd.DataFrame] = None) -> pd.DataFrame:
        pass

    def fit_transform(self, X, y=None):
        self.fit(X, y)
        return self.transform(X)



import numpy as np
import pandas as pd

from giottotime.feature_creation.index_dependent_features.tda_features import (
    AmplitudeFeature,
)


def test_correct_amplitude_computation():
    np.random.seed(0)

    output_name = "ignored"
    amplitude_feature = AmplitudeFeature(output_name=output_name)
    df = pd.DataFrame(np.random.randint(0, 100, size=(30, 1)), columns=["ignored"])

    persistence_diagrams = amplitude_feature._compute_persistence_diagrams(df)
    amplitudes = amplitude_feature._calculate_amplitude_feature(persistence_diagrams)

    expected_amplitudes = np.array(
        [
            [0.42390011],
            [0.42392732],
            [0.48618924],
            [0.46564008],
            [0.46508716],
            [0.33761475],
            [0.33761475],
            [0.33680451],
            [0.29547906],
            [0.30680885],
            [0.29692516],
            [0.29494336],
            [0.35830464],
            [0.35827269],
            [0.35806485],
            [0.37202706],
            [0.3731072],
        ]
    )

    assert np.allclose(expected_amplitudes, amplitudes)


def test_correct_amplitude_feature():
    np.random.seed(0)

    output_name = "amplitude_feature"
    amplitude_feature = AmplitudeFeature(output_name=output_name)
    df = pd.DataFrame(np.random.randint(0, 100, size=(30, 1)), columns=["old_name"])

    amplitudes = amplitude_feature.fit_transform(df)
    expected_amplitudes = pd.DataFrame.from_dict(
        {
            output_name: [
                0.423900,
                0.423900,
                0.423927,
                0.423927,
                0.486189,
                0.486189,
                0.465640,
                0.465640,
                0.465087,
                0.465087,
                0.337615,
                0.337615,
                0.337615,
                0.337615,
                0.336805,
                0.336805,
                0.295479,
                0.295479,
                0.306809,
                0.306809,
                0.296925,
                0.296925,
                0.294943,
                0.294943,
                0.358305,
                0.358305,
                0.358273,
                0.358065,
                0.372027,
                0.373107,
            ]
        }
    )
    expected_amplitudes.index = df.index

    assert np.allclose(expected_amplitudes.values, amplitudes.values)
    assert np.array_equal(expected_amplitudes.index, amplitudes.index)
    assert expected_amplitudes.columns == amplitudes.columns
